package com.kms.service;

import com.j256.ormlite.dao.Dao;
import com.kms.crosscut.SoapFaultHelper;
import com.kms.dto.EmailReq;
import com.kms.dto.payment.PaymentHistoryResp;
import com.kms.model.Payment;
import com.kms.service.helper.PaymentHistoryCacheHelper;
import com.kms.util.PdfUtil;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import lombok.RequiredArgsConstructor;

import javax.activation.DataHandler;
import javax.activation.FileDataSource;
import javax.annotation.Resource;
import javax.jws.HandlerChain;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.xml.bind.annotation.XmlMimeType;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.soap.MTOM;
import java.io.IOException;
import java.io.StringWriter;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

@WebService
@RequiredArgsConstructor
@HandlerChain(file = "handlers.xml")
@MTOM(threshold = 512)
public class PaymentHistoryService {

    @Resource
    private WebServiceContext context;

    private final Dao<Payment,Integer> paymentDao;

    private static final Configuration cfg = new Configuration(Configuration.VERSION_2_3_30);

    static {
        cfg.setClassForTemplateLoading(PaymentHistoryService.class, "/");
        cfg.setDefaultEncoding("UTF-8");
        cfg.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
    }


    @WebMethod
    public List<PaymentHistoryResp> getPaymentHistory(@WebParam(name = "initiatorId") String initiatorId,
                                                      @WebParam(name = "sendMail") boolean sendMail,
                                                      @WebParam(name = "mailData") EmailReq emailReq) {
        MessageContext messageContext = context.getMessageContext();

        String cacheKey = PaymentHistoryCacheHelper.checkForHistoryCacheAvailability(initiatorId);
        if (!cacheKey.isEmpty()) {
            messageContext.put("returnFromCache", true);
            messageContext.put("cacheKey", cacheKey);
            return null;
        }

        try {
            Payment filter = Payment.builder()
                    .initiatorId(initiatorId)
                    .paid(true)
                    .build();
            List<Payment> payments = paymentDao.queryForMatching(filter);

            if (sendMail && emailReq != null) {
                String html = constructHistoryHtml(payments);

                EmailService.send(emailReq.getRecipient(), emailReq.getCc(),
                        "KMS Payment History", html);
            }

            cacheKey = PaymentHistoryCacheHelper.formHistoryCacheKey(initiatorId);
            messageContext.put("cacheResponse", true);
            messageContext.put("cacheKey", cacheKey);

            return payments.stream()
                    .filter(el -> !(el.getPaymentInitTime() == null || el.getPaymentPaidTime() == null))
                    .map(el -> {
                        String initTime = Optional.of(el.getPaymentInitTime())
                                .orElseGet(() -> new Timestamp(0))
                                .toString(),
                            paidTime = Optional.of(el.getPaymentPaidTime())
                                .orElseGet(() -> new Timestamp(0))
                                .toString();

                        return new PaymentHistoryResp(initTime, paidTime, el.getAmount(), el.getDescription());
                    })
                    .collect(Collectors.toList());
        } catch (SQLException | TemplateException | IOException e) {
            return null;
        }
    }

    @WebMethod
    public boolean sendPaymentHistoryMail(@WebParam(name = "initiatorId") String initiatorId,
                                          @WebParam(name = "mailData") EmailReq emailReq) {
        if (emailReq == null)
            return false;

        try {
            Payment filter = Payment.builder()
                    .initiatorId(initiatorId)
                    .paid(true)
                    .build();
            List<Payment> payments = paymentDao.queryForMatching(filter);
            EmailService.send(emailReq.getRecipient(), emailReq.getCc(), "KMS Payment History",
                    constructHistoryHtml(payments));
            return true;
        } catch (SQLException | TemplateException | IOException e) {
            return false;
        }
    }

    @WebMethod
    @XmlMimeType("application/pdf")
    public DataHandler getPaymentHistoryPdf(@WebParam(name = "initiatorId") String initiatorId) {
        String filename = formPdfFilename(initiatorId);
        String filePath = PdfUtil.getPdfFolder()+filename;

        // caching
        if (PdfUtil.checkPdfExists(filePath))
            return getPdfDataHandler(filePath);

        try {
            Payment filter = Payment.builder()
                    .initiatorId(initiatorId)
                    .paid(true)
                    .build();
            List<Payment> payments = paymentDao.queryForMatching(filter);

            String html = constructHistoryHtml(payments);
            PdfUtil.writePdfToFile(html, filename);

            if (!PdfUtil.checkPdfExists(filename)) {
                System.out.println("ERROR: PDF DOES NOT EXIST AFTER WRITE. PATH: " + filePath);
                SoapFaultHelper.throwFault("UnexpectedError", "Unexpected error has happened.");
            }
            return getPdfDataHandler(filePath);
        } catch (SQLException | TemplateException | IOException e) {
            System.out.println(e.getMessage());
            SoapFaultHelper.throwFault("UnexpectedError", "Unexpected error has happened.");
        }

        return null;
    }

    private DataHandler getPdfDataHandler(String filePath) {
        return new DataHandler(new FileDataSource(filePath));
    }

    private String formPdfFilename(String initiatorId) {
        return "payment-history-"+initiatorId+".pdf";
    }

    private String constructHistoryHtml(List<Payment> payments) throws IOException, TemplateException {
        Template template = cfg.getTemplate("history.ftl");
        Map<String, Object> dataModel = new HashMap<>();
        dataModel.put("payments", payments);

        StringWriter stringWriter = new StringWriter();
        template.process(dataModel, stringWriter);
        return stringWriter.toString();
    }

}
